@@ -139,6 +139,7 @@ int iso_node_xinfo_dispose_cloners(int flag)
         next = assoc->next;
         free((char *) assoc);
     }
+    iso_xinfo_cloner_list= NULL;
     return(1);
 }
 
@@ -344,9 +345,9 @@ const char *iso_error_to_msg(int errcode)
     case ISO_WRONG_PVD:
         return "Wrong or damaged Primary Volume Descriptor";
     case ISO_WRONG_RR:
-        return "Wrong or damaged RR entry";
+        return "Wrong or damaged Rock Ridge entry";
     case ISO_UNSUPPORTED_RR:
-        return "Unsupported RR feature";
+        return "Unsupported Rock Ridge feature";
     case ISO_WRONG_ECMA119:
         return "Wrong or damaged ECMA-119";
     case ISO_UNSUPPORTED_ECMA119:
@@ -360,9 +361,9 @@ const char *iso_error_to_msg(int errcode)
     case ISO_UNSUPPORTED_SUSP:
         return "Unsupported SUSP feature";
     case ISO_WRONG_RR_WARN:
-        return "Error on a RR entry that can be ignored";
+        return "Error on a Rock Ridge entry that can be ignored";
     case ISO_SUSP_UNHANDLED:
-        return "Error on a RR entry that can be ignored";
+        return "Unhandled SUSP entry";
     case ISO_SUSP_MULTIPLE_ER:
         return "Multiple ER SUSP entries found";
     case ISO_UNSUPPORTED_VD:
@@ -500,6 +501,8 @@ const char *iso_error_to_msg(int errcode)
         return "Symbolic link cannot be resolved";
     case ISO_DEEP_SYMLINK:
         return "Too many chained symbolic links";
+    case ISO_BAD_ISO_FILETYPE:
+        return "Unrecognized file type in ISO image";
     default:
         return "Unknown error";
     }
