@@ -55,6 +55,8 @@ void iso_node_ref(IsoNode *node)
  */
 void iso_node_unref(IsoNode *node)
 {
+    if (node == NULL)
+        return;
     if (--node->refcount == 0) {
         switch (node->type) {
         case LIBISO_DIR:
@@ -1046,6 +1048,7 @@ void iso_node_set_sort_weight(IsoNode *node, int w)
         }
     } else if (node->type == LIBISO_FILE) {
         ((IsoFile*)node)->sort_weight = w;
+        ((IsoFile*)node)->explicit_weight = 1;
     }
 }
 
@@ -1428,6 +1431,8 @@ int iso_node_new_file(char *name, IsoStream *stream, IsoFile **file)
     new->node.type = LIBISO_FILE;
     new->node.name = name;
     new->node.mode = S_IFREG;
+    new->from_old_session = 0;
+    new->explicit_weight = 0;
     new->sort_weight = 0;
     new->stream = stream;
 
@@ -1725,6 +1730,7 @@ int attr_enlarge_list(char ***names, size_t **value_lengths, char ***values,
                         bit2= delete the given names rather than overwrite
                               their content
                         bit4= do not overwrite value of empty name
+                        bit5= do not overwrite isofs attributes 
                         bit15= release memory and return 1
 */
 static
@@ -1773,6 +1779,8 @@ int iso_node_merge_xattr(IsoNode *node, size_t num_attrs, char **names,
             continue;
         if (names[i][0] == 0 && (flag & 16))
             continue;
+        if ((flag & 32) && strncmp(names[i], "isofs.", 6) == 0)
+            continue;
         for (j = 0; j < *m_num_attrs; j++) {
             if ((*m_names)[j] == NULL)
                 continue;
@@ -1818,6 +1826,8 @@ int iso_node_merge_xattr(IsoNode *node, size_t num_attrs, char **names,
                 continue;
             if (names[i][0] == 0 && (flag & 16))
                 continue;
+            if ((flag & 32) && strncmp(names[i], "isofs.", 6) == 0)
+                continue;
             for (j = 0; j < *m_num_attrs; j++) {
                 if ((*m_names)[j] == NULL)
                     continue;
@@ -1868,13 +1878,13 @@ int iso_node_set_attrs(IsoNode *node, size_t num_attrs, char **names,
         for (i = 0; i < num_attrs; i++)
             if (strncmp(names[i], "user.", 5) != 0 && names[i][0] != 0) 
                 return ISO_AAIP_NON_USER_NAME;  
-
-    if ((flag & (2 | 4)) || !(flag & 8)) {
+    if ((flag & (2 | 4 | 16)) || !(flag & 8)) {
         /* Merge old and new lists */
         ret = iso_node_merge_xattr(
                   node, num_attrs, names, value_lengths, values,
                   &m_num, &m_names, &m_value_lengths, &m_values,
-                  (flag & 4) | (!(flag & 2)) | ((!(flag & 1)) << 4));
+                  (flag & 4) | (!(flag & 2)) | ((!(flag & 1)) << 4) |
+                  ((flag & 16) << 1));
         if (ret < 0)
             goto ex;
         num_attrs = m_num;
@@ -2559,6 +2569,7 @@ int iso_node_set_ino(IsoNode *node, ino_t ino, int flag)
         ret = iso_stream_set_image_ino(file->stream, ino, 0);
         if (ret < 0 || ret == 1)
             return ret;
+        /* ret == 0 means that the stream is not from loaded ISO image */
 
     } else if (node->type == LIBISO_SYMLINK) {
         symlink = (IsoSymlink *) node;
@@ -2749,6 +2760,8 @@ int iso_node_cmp_ino(IsoNode *n1, IsoNode *n2, int flag)
 }
 
 
+/* @param flag bit0= delete isofs.cx rather than setting it
+*/
 int iso_file_set_isofscx(IsoFile *file, unsigned int checksum_index,
                          int flag)
 {
@@ -2758,9 +2771,14 @@ int iso_file_set_isofscx(IsoFile *file, unsigned int checksum_index,
     char *valuept;
     int i, ret;
 
+    valuept= (char *) value;
+    if (flag & 1) {
+        ret = iso_node_set_attrs((IsoNode *) file, (size_t) 1,
+                             &names, value_lengths, &valuept, 4 | 8);
+        return ret;
+    }
     for(i = 0; i < 4; i++)
         value[3 - i] = (checksum_index >> (8 * i)) & 0xff;
-    valuept= (char *) value;
     ret = iso_node_set_attrs((IsoNode *) file, (size_t) 1,
                              &names, value_lengths, &valuept, 2 | 8);
     return ret;
