@@ -1,6 +1,6 @@
 /*
  * Copyright (c) 2008 Vreixo Formoso
- * Copyright (c) 2010 - 2012 Thomas Schmitt
+ * Copyright (c) 2010 - 2013 Thomas Schmitt
  *
  * This file is part of the libisofs project; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License version 2 
@@ -27,6 +27,7 @@
 #include <sys/types.h>
 #include <sys/stat.h>
 #include <unistd.h>
+#include <fcntl.h>
 
 /* for gettimeofday() */
 #include <sys/time.h>
@@ -71,21 +72,25 @@ static int precompute_gpt(Ecma119Image *t);
 /*
  * @param flag bit0= img_blocks is start address rather than end address:
                      do not subtract 1
+               bit1= img_blocks is counted in 512-byte units rather than 2 KiB
  */
 static
-void iso_compute_cyl_head_sec(uint32_t *img_blocks, int hpc, int sph,
+void iso_compute_cyl_head_sec(uint64_t img_blocks, int hpc, int sph,
                               uint32_t *end_lba, uint32_t *end_sec,
                               uint32_t *end_head, uint32_t *end_cyl, int flag)
 {
-    uint32_t secs;
+    uint64_t secs;
 
-    /* Partition table unit is 512 bytes per sector, ECMA-119 unit is 2048 */
-    if (*img_blocks >= 0x40000000)
-      *img_blocks = 0x40000000 - 1;        /* truncate rather than roll over */
-    if (flag & 1)
-        secs = *end_lba = *img_blocks * 4;            /* first valid 512-lba */
+    if(flag & 2)
+        secs = img_blocks;
     else
-        secs = *end_lba = *img_blocks * 4 - 1;         /* last valid 512-lba */
+        secs = img_blocks * 4;
+    if (secs > (uint64_t) 0xfffffffc)
+      secs = 0xfffffffc;                   /* truncate rather than roll over */
+    if (flag & 1)
+        *end_lba = secs;                              /* first valid 512-lba */
+    else
+        secs = *end_lba = secs - 1;                    /* last valid 512-lba */
     *end_cyl = secs / (sph * hpc);
     secs -= *end_cyl * sph * hpc;
     *end_head = secs / sph;
@@ -121,10 +126,18 @@ static uint32_t compute_partition_size(char *disk_path, uint32_t *size,
 */
 int iso_compute_append_partitions(Ecma119Image *t, int flag)
 {
-    int ret, i, sa_type;
+    int ret, i, sa_type, cyl_align, cyl_size = 0;
     uint32_t pos, size, add_pos = 0;
 
     sa_type = (t->system_area_options >> 2) & 0x3f;
+    cyl_align = (t->system_area_options >> 8) & 0x3;
+    if (sa_type == 0 && cyl_align == 3) {
+        cyl_size = t->partition_heads_per_cyl * t->partition_secs_per_head;
+        if (cyl_size % 4)
+            cyl_size = 0;
+        else
+            cyl_size /= 4;
+    }
     pos = (t->vol_space_size + t->ms_block);
     for (i = 0; i < ISO_MAX_PARTITIONS; i++) {
         if (t->appended_partitions[i] == NULL)
@@ -134,10 +147,20 @@ int iso_compute_append_partitions(Ecma119Image *t, int flag)
         ret = compute_partition_size(t->appended_partitions[i], &size, 0);
         if (ret < 0)
             return ret;
-        if (sa_type == 3 && (pos % ISO_SUN_CYL_SIZE))
+        add_pos = 0;
+        if (sa_type == 3 && (pos % ISO_SUN_CYL_SIZE)) {
             add_pos = ISO_SUN_CYL_SIZE - (pos % ISO_SUN_CYL_SIZE);
+        } else if (cyl_size > 0 && (pos % cyl_size)) {
+            add_pos = cyl_size - (pos % cyl_size);
+        }
         t->appended_part_prepad[i] = add_pos;
         t->appended_part_start[i] = pos + add_pos;
+
+        if (cyl_size > 0 && (size % cyl_size)) {
+            /* Obey cylinder alignment (missing data will be written as
+               zeros by iso_write_partition_file()) */
+            size += cyl_size - (size % cyl_size);
+        }
         t->appended_part_size[i] = size;
         pos += add_pos + size;
         t->total_size += (add_pos + size) * 2048;
@@ -146,10 +169,12 @@ int iso_compute_append_partitions(Ecma119Image *t, int flag)
 }
 
 
-/* Note: partition_offset and partition_size are counted in 2048 blocks
+/* @param flag
+                bit1= partition_offset and partition_size are counted in
+                      blocks of 512 rather than 2048
  */
 static int write_mbr_partition_entry(int partition_number, int partition_type,
-                  uint32_t partition_offset, uint32_t partition_size,
+                  uint64_t partition_offset, uint64_t partition_size,
                   int sph, int hpc, uint8_t *buf, int flag)
 {
     uint8_t *wpt;
@@ -159,10 +184,12 @@ static int write_mbr_partition_entry(int partition_number, int partition_type,
     int i;
 
     after_end = partition_offset + partition_size;
-    iso_compute_cyl_head_sec(&partition_offset, hpc, sph,
-                           &start_lba, &start_sec, &start_head, &start_cyl, 1);
-    iso_compute_cyl_head_sec(&after_end, hpc, sph,
-                             &end_lba, &end_sec, &end_head, &end_cyl, 0);
+    iso_compute_cyl_head_sec((uint64_t) partition_offset, hpc, sph,
+                             &start_lba, &start_sec, &start_head, &start_cyl,
+                             1 | (flag & 2));
+    iso_compute_cyl_head_sec((uint64_t) after_end, hpc, sph,
+                             &end_lba, &end_sec, &end_head, &end_cyl,
+                             (flag & 2));
     wpt = buf + 446 + (partition_number - 1) * 16;
 
     /* Not bootable */
@@ -225,7 +252,7 @@ int make_grub_msdos_label(uint32_t img_blocks, int sph, int hpc,
     uint32_t end_lba, end_sec, end_head, end_cyl;
     int i;
 
-    iso_compute_cyl_head_sec(&img_blocks, hpc, sph,
+    iso_compute_cyl_head_sec((uint64_t) img_blocks, hpc, sph,
                              &end_lba, &end_sec, &end_head, &end_cyl, 0);
 
     /* 1) Zero-fill 446-510 */
@@ -289,9 +316,9 @@ int iso_offset_partition_start(uint32_t img_blocks, uint32_t partition_offset,
     uint32_t start_lba, start_sec, start_head, start_cyl;
     int i;
 
-    iso_compute_cyl_head_sec(&partition_offset, hpc, sph,
+    iso_compute_cyl_head_sec((uint64_t) partition_offset, hpc, sph,
                            &start_lba, &start_sec, &start_head, &start_cyl, 1);
-    iso_compute_cyl_head_sec(&img_blocks, hpc, sph,
+    iso_compute_cyl_head_sec((uint64_t) img_blocks, hpc, sph,
                              &end_lba, &end_sec, &end_head, &end_cyl, 0);
     wpt = buf + 446;
 
@@ -672,7 +699,8 @@ static int write_sun_partition_entry(int partition_number,
  */
 static int make_sun_disk_label(Ecma119Image *t, uint8_t *buf, int flag)
 {
-    int ret;
+    int ret, i;
+    uint64_t blk;
 
     /* Bytes 512 to 32767 may come from image or external file */
     memset(buf, 0, 512);
@@ -714,6 +742,16 @@ static int make_sun_disk_label(Ecma119Image *t, uint8_t *buf, int flag)
     /* 508 - 509 |     0xdabe | Magic Number */
     iso_msb(buf + 508, 0xdabe, 2);
 
+    if (t->sparc_core_src != NULL) {
+        /* May be used for grub-sparc. */
+        blk= ((uint64_t) t->sparc_core_src->sections[0].block) *
+             (uint64_t) 2048; 
+        for (i = 0; i < 8; i++)
+            buf[Libisofs_grub2_sparc_patch_adr_poS + i] = blk >> ((7 - i) * 8);
+        iso_msb(buf + Libisofs_grub2_sparc_patch_size_poS,
+                t->sparc_core_src->sections[0].size, 4);
+    }
+
     /* Set partition 1 to describe ISO image and compute checksum */
     t->appended_part_start[0] = 0;
     t->appended_part_size[0] = t->curblock;
@@ -775,7 +813,7 @@ int iso_quick_gpt_entry(Ecma119Image *t,
 
 
 int iso_quick_mbr_entry(Ecma119Image *t,
-                        uint32_t start_block, uint32_t block_count,
+                        uint64_t start_block, uint64_t block_count,
                         uint8_t type_byte, uint8_t status_byte,
                         int desired_slot)
 {
@@ -861,7 +899,7 @@ static int iso_write_apm_entry(Ecma119Image *t, int apm_block_size,
     else
         block_fac = 2048 / apm_block_size;
         
-    memset(buf, apm_block_size, 0);
+    memset(buf, 0, apm_block_size);
     wpt = buf;
 
     /* Signature */
@@ -1096,10 +1134,11 @@ static int iso_write_mbr(Ecma119Image *t, uint32_t img_blocks, uint8_t *buf)
 
     /* <<< Dummy mock-up */
     if (t->mbr_req_count <= 0) {
-        ret = iso_quick_mbr_entry(t, 0, 0, 0xee, 0, 0);
+        ret = iso_quick_mbr_entry(t, (uint64_t) 0, (uint64_t) 0, 0xee, 0, 0);
         if (ret < 0)
             return ret;
-        ret = iso_quick_mbr_entry(t, 100, 0, 0x0c, 0x80, 1);
+        ret = iso_quick_mbr_entry(t, ((uint64_t) 100) * 4, (uint64_t) 0,
+                                  0x0c, 0x80, 1);
         if (ret < 0)
             return ret;
     }
@@ -1128,7 +1167,7 @@ static int iso_write_mbr(Ecma119Image *t, uint32_t img_blocks, uint8_t *buf)
             t->mbr_req[i]->block_count = t->mbr_req[i + 1]->start_block -
                                          t->mbr_req[i]->start_block;
         else
-            t->mbr_req[i]->block_count = img_blocks -
+            t->mbr_req[i]->block_count = ((uint64_t) img_blocks) * 4 -
                                          t->mbr_req[i]->start_block;
     }
 
@@ -1165,7 +1204,7 @@ static int iso_write_mbr(Ecma119Image *t, uint32_t img_blocks, uint8_t *buf)
         ret = write_mbr_partition_entry(i + 1, (int) t->mbr_req[q]->type_byte,
                   t->mbr_req[q]->start_block, t->mbr_req[q]->block_count,
                   t->partition_secs_per_head, t->partition_heads_per_cyl,
-                  buf, 0);
+                  buf, 2);
         if (ret < 0)
             return ret;
         buf[446 + i * 16] = t->mbr_req[q]->status_byte;
@@ -1279,7 +1318,7 @@ int iso_write_gpt_header_block(Ecma119Image *t, uint32_t img_blocks,
     }
 
     /* CRC-32 of this header while head_crc is 0 */
-    crc = iso_crc32_gpt((unsigned char *) buf, 512, 0); 
+    crc = iso_crc32_gpt((unsigned char *) buf, 92, 0); 
     wpt = ((char *) buf) + 16;
     iso_lsb_to_buf(&wpt, crc, 4, 0);
 
@@ -1398,6 +1437,8 @@ int iso_write_system_area(Ecma119Image *t, uint8_t *buf)
     int first_partition = 1, last_partition = 4, apm_flag, part_type;
     int gpt_count = 0, gpt_idx[128], apm_count = 0;
     uint32_t img_blocks;
+    uint64_t blk;
+    uint8_t *wpt;
 
     if ((t == NULL) || (buf == NULL)) {
         return ISO_NULL_POINTER;
@@ -1545,7 +1586,8 @@ int iso_write_system_area(Ecma119Image *t, uint8_t *buf)
             return ISO_ASSERT_FAILURE;
         if (t->partition_offset == 0) {
             /* Re-write partion entry 1 : start at 0, type Linux */
-            ret = write_mbr_partition_entry(1, 0x83, 0, img_blocks,
+            ret = write_mbr_partition_entry(1, 0x83,
+                        (uint64_t) 0, (uint64_t) img_blocks,
                         t->partition_secs_per_head, t->partition_heads_per_cyl,
                         buf, 0);
             if (ret < 0)
@@ -1560,7 +1602,7 @@ int iso_write_system_area(Ecma119Image *t, uint8_t *buf)
         img_blocks = t->curblock;                  /* value might be altered */
         ret = iso_offset_partition_start(img_blocks, t->partition_offset,
                                          t->partition_secs_per_head,
-                                         t->partition_heads_per_cyl, buf, 1);
+                                         t->partition_heads_per_cyl, buf, 0);
         if (ret != ISO_SUCCESS) /* error should never happen */
             return ISO_ASSERT_FAILURE;
     }
@@ -1580,7 +1622,8 @@ int iso_write_system_area(Ecma119Image *t, uint8_t *buf)
                         buf, t->appended_partitions[i][0] == 0);
         } else {
             ret = write_mbr_partition_entry(i + 1, t->appended_part_types[i],
-                        t->appended_part_start[i], t->appended_part_size[i],
+                        (uint64_t) t->appended_part_start[i],
+                        (uint64_t) t->appended_part_size[i],
                         t->partition_secs_per_head, t->partition_heads_per_cyl,
                         buf, 0);
         }
@@ -1588,6 +1631,15 @@ int iso_write_system_area(Ecma119Image *t, uint8_t *buf)
             return ret;
     }
 
+    if (sa_type == 0 && (t->system_area_options & 0x4000) && !do_isohybrid) {
+        /* Patch MBR for GRUB2 */
+        blk = t->bootsrc[0]->sections[0].block * 4 +
+                                                Libisofs_grub2_mbr_patch_offsT;
+        wpt = buf + Libisofs_grub2_mbr_patch_poS;
+        for (i = 0; i < 8; i++)
+             wpt[i] = blk >> (i * 8);
+    }
+
     return ISO_SUCCESS;
 }
 
@@ -1663,7 +1715,7 @@ int iso_align_isohybrid(Ecma119Image *t, int flag)
         iso_msgs_submit(0, msg, 0, "NOTE", 0);
     }
 
-    if (always_align >= 2)
+    if (always_align == 2)
         {ret = ISO_SUCCESS; goto ex;}
 
     cylsize = 0;
@@ -1834,12 +1886,13 @@ void iso_random_uuid(Ecma119Image *t, uint8_t uuid[16])
         0xee, 0x29, 0x9d, 0xfc, 0x65, 0xcc, 0x7c, 0x40,
         0x92, 0x61, 0x5b, 0xcd, 0x6f, 0xed, 0x08, 0x34
     };
+    static uint8_t uuid_urandom[16];
     uint32_t rnd, salt;
     struct timeval tv;
     struct timezone tz;
     pid_t pid;
-    static int counter = 0;
-    int i;
+    static int counter = 0, use_urandom = 0;
+    int i, ret, fd;
 #endif
 
 #ifdef Libisofs_with_uuid_generatE
@@ -1850,6 +1903,32 @@ void iso_random_uuid(Ecma119Image *t, uint8_t uuid[16])
 
 #else
 
+    /* First try /dev/urandom.
+       (Weakening the result by 8 bit saves a lot of pool entropy.)
+    */
+    if ((counter & 0xff) == 0) {
+        fd = open("/dev/urandom", O_RDONLY);
+        if (fd == -1)
+            goto fallback;
+        ret = read(fd, uuid_urandom, 16);
+        if (ret != 16) {
+            close(fd);
+            goto fallback;
+        }
+        /* Mark as UUID version 4 */
+        uuid_urandom[7] = (uuid_urandom[7] & 0x0f) | 0x40;
+        uuid_urandom[8] = (uuid_urandom[8] & 0x3f) | 0x80;
+        close(fd);
+        use_urandom = 1;
+    }
+    if (!use_urandom)
+        goto fallback;
+    memcpy(uuid, uuid_urandom, 16);
+    uuid[9] ^= counter & 0xff;
+    counter++;
+    return;
+
+fallback:;
     pid = getpid();
     salt = iso_crc32_gpt((unsigned char *) t, sizeof(Ecma119Image), 0) ^ pid; 
 
@@ -1865,7 +1944,7 @@ void iso_random_uuid(Ecma119Image *t, uint8_t uuid[16])
         u[i] = (salt >> (8 * i)) & 0xff;
     for (i = 0; i < 2; i++)
         u[4 + i] = (pid >> (8 * i)) & 0xff;
-    u[6] = ((salt >> 8) | (pid >> 16)) & 0xff;
+    u[6] = ((salt >> 8) ^ (pid >> 16)) & 0xff;
     rnd = ((0xffffff & tv.tv_sec) << 8) |
           (((tv.tv_usec >> 16) ^ (salt & 0xf0)) & 0xff);
     u[9] ^= counter & 0xff;
@@ -1882,21 +1961,6 @@ void iso_random_uuid(Ecma119Image *t, uint8_t uuid[16])
 }
 
 
-void iso_random_8byte(Ecma119Image *t, uint8_t result[8])
-{
-    uint8_t uuid[16];
-    int i;
-    
-    iso_random_uuid(t, uuid);
-    for (i = 0; i < 8; i++) {
-        if (i == 1) 
-            result[i] = uuid[9]; /* The intra-process counter */
-        else 
-            result[i] = uuid[i] ^ uuid[i + 8];
-    }
-}
-
-
 /* Probably already called by tail_writer_compute_data_blocks via
    iso_align_isohybrid
 */
@@ -2084,7 +2148,7 @@ tampered_head:;
 
     /* Compute new header CRC */
     memset(new_head + 16, 0, 4);
-    crc = iso_crc32_gpt((unsigned char *) new_head, 512, 0); 
+    crc = iso_crc32_gpt((unsigned char *) new_head, 92, 0); 
     iso_lsb(new_head + 16, crc, 4);
 
     /* Copy GPT entries */
@@ -2195,7 +2259,7 @@ static int partprepend_writer_compute_data_blocks(IsoImageWriter *writer)
         if (t->prep_partition != NULL || t->fat || will_have_gpt ||
             t->mbr_req_count > 0)
             return ISO_BOOT_MBR_OVERLAP;
-        ret = iso_quick_mbr_entry(t, (uint32_t) 0, (uint32_t) 0,
+        ret = iso_quick_mbr_entry(t, (uint64_t) 0, (uint64_t) 0,
                                   0x96, 0x80, 0);
         if (ret < 0)
             return ret;
@@ -2210,21 +2274,25 @@ static int partprepend_writer_compute_data_blocks(IsoImageWriter *writer)
     }
     if (t->prep_part_size > 0 || t->fat || will_have_gpt) {
         /* Protecting MBR entry for ISO start or whole ISO */
-        ret = iso_quick_mbr_entry(t, (uint32_t) t->partition_offset,
-                              (uint32_t) 0, will_have_gpt ? 0xee : 0xcd, 0, 0);
+        ret = iso_quick_mbr_entry(t, will_have_gpt ? (uint64_t) 1 :
+                                  ((uint64_t) t->partition_offset) * 4,
+                                  (uint64_t) 0,
+                                  will_have_gpt ? 0xee : 0xcd, 0, 0);
         if (ret < 0)
             return ret;
     }
     if (t->prep_part_size > 0) {
-        ret = iso_quick_mbr_entry(t, t->curblock, t->prep_part_size, 0x41,
-                                  0, 0);
+        ret = iso_quick_mbr_entry(t, ((uint64_t) t->curblock) * 4,
+                                  ((uint64_t) t->prep_part_size) * 4,
+                                  0x41, 0, 0);
         if (ret < 0)
             return ret;
         t->curblock += t->prep_part_size;
     }
     if (t->prep_part_size > 0 || t->fat) {
         /* FAT partition or protecting MBR entry for ISO end */
-        ret = iso_quick_mbr_entry(t, (uint32_t) t->curblock, (uint32_t) 0,
+        ret = iso_quick_mbr_entry(t, ((uint64_t) t->curblock) * 4,
+                                  (uint64_t) 0,
                                   t->fat ? 0x0c : 0xcd, 0, 0);
         if (ret < 0)
             return ret;
