@@ -478,12 +478,12 @@ int iso_tree_remove_exclude(IsoImage *image, const char *path)
         return ISO_NULL_POINTER;
     }
 
-    for (i = 0; i < image->nexcludes; ++i) {
+    for (i = 0; (int) i < image->nexcludes; ++i) {
         if (strcmp(image->excludes[i], path) == 0) {
             /* exclude found */
             free(image->excludes[i]);
             --image->nexcludes;
-            for (j = i; j < image->nexcludes; ++j) {
+            for (j = i; (int) j < image->nexcludes; ++j) {
                 image->excludes[j] = image->excludes[j+1]; 
             }
             image->excludes = realloc(image->excludes, image->nexcludes * 
@@ -761,11 +761,16 @@ int iso_add_dir_src_rec(IsoImage *image, IsoDir *parent, IsoFileSource *dir)
 
     ret = iso_file_source_open(dir);
     if (ret < 0) {
-        char *path = iso_file_source_get_path(dir);
+        path = iso_file_source_get_path(dir);
         /* instead of the probable error, we throw a sorry event */
-        ret = iso_msg_submit(image->id, ISO_FILE_CANT_ADD, ret, 
-                             "Can't open dir %s", path);
-        free(path);
+	if (path != NULL) {
+            ret = iso_msg_submit(image->id, ISO_FILE_CANT_ADD, ret, 
+                                 "Can't open dir %s", path);
+            free(path);
+        } else {
+            ret = iso_msg_submit(image->id, ISO_NULL_POINTER, ret,
+                           "Can't open dir. NULL pointer caught as dir name");
+        }
         return ret;
     }
 
@@ -785,6 +790,11 @@ int iso_add_dir_src_rec(IsoImage *image, IsoDir *parent, IsoFileSource *dir)
         }
 
         path = iso_file_source_get_path(file);
+        if (path == NULL) {
+            ret = iso_msg_submit(image->id, ISO_NULL_POINTER, ret, 
+                                 "NULL pointer caught as file path");
+            return ret;
+        }
         name = strrchr(path, '/') + 1;
 
         if (image->follow_symlinks) {
@@ -793,6 +803,8 @@ int iso_add_dir_src_rec(IsoImage *image, IsoDir *parent, IsoFileSource *dir)
             ret = iso_file_source_lstat(file, &info);
         }
         if (ret < 0) {
+            ret = iso_msg_submit(image->id, ISO_FILE_CANT_ADD, ret,
+                                 "Error when adding file %s", path);
             goto dir_rec_continue;
         }
 
@@ -847,7 +859,7 @@ int iso_add_dir_src_rec(IsoImage *image, IsoDir *parent, IsoFileSource *dir)
         ret = iso_dir_insert(parent, new, pos, replace);
         if (ret < 0) {
             iso_node_unref(new);
-            if (ret != ISO_NODE_NAME_NOT_UNIQUE) {
+            if (ret != (int) ISO_NODE_NAME_NOT_UNIQUE) {
                 /* error */
                 goto dir_rec_continue;
             } else {
@@ -1186,3 +1198,84 @@ int iso_tree_clone(IsoNode *node,
     return ret;
 }
 
+
+int iso_tree_resolve_symlink(IsoImage *img, IsoSymlink *sym, IsoNode **res,
+                             int *depth, int flag)
+{
+    IsoDir *cur_dir = NULL;
+    IsoNode *n, *resolved_node;
+    char *dest, *dest_start, *dest_end;
+    int ret = 0;
+    unsigned int comp_len, dest_len;
+
+    dest = sym->dest;
+    dest_len = strlen(dest);
+
+    if (dest[0] == '/') {
+
+        /* ??? How to resolve absolute links without knowing the
+               path of the future mount point ?
+           ??? Would it be better to throw error ? 
+           I can only assume that it gets mounted at / during some stage
+           of booting.
+        */;
+
+        cur_dir = img->root;
+        dest_end = dest;
+    } else {
+        cur_dir = sym->node.parent;
+        if (cur_dir == NULL)
+            cur_dir = img->root;
+        dest_end = dest - 1;
+    }
+
+    while (dest_end < dest + dest_len) {
+        dest_start = dest_end + 1;
+        dest_end = strchr(dest_start, '/');
+        if (dest_end == NULL)
+            dest_end = dest_start + strlen(dest_start);
+        comp_len = dest_end - dest_start;
+        if (comp_len == 0 || (comp_len == 1 && dest_start[0] == '.'))
+    continue;
+        if (comp_len == 2 && dest_start[0] == '.' && dest_start[1] == '.') {
+            cur_dir = cur_dir->node.parent;
+            if (cur_dir == NULL) /* link shoots over root */
+                return ISO_DEAD_SYMLINK;
+    continue;
+        }
+
+        /* Search node in cur_dir */
+        for (n = cur_dir->children; n != NULL; n = n->next)
+            if (strncmp(dest_start, n->name, comp_len) == 0 &&
+                strlen(n->name) == comp_len)
+        break;
+        if (n == NULL)
+            return ISO_DEAD_SYMLINK;
+
+        if (n->type == LIBISO_DIR) {
+            cur_dir = (IsoDir *) n;
+        } else if (n->type == LIBISO_SYMLINK) {
+            if (*depth >= LIBISO_MAX_LINK_DEPTH)
+                return ISO_DEEP_SYMLINK;
+            (*depth)++;
+            ret = iso_tree_resolve_symlink(img, (IsoSymlink *) n,
+                                           &resolved_node, depth, 0);
+            if (ret < 0)
+                return ret;
+            if (resolved_node->type != LIBISO_DIR) {
+                n = resolved_node;
+                goto leaf_type;
+            }
+            cur_dir = (IsoDir *) resolved_node;
+        } else {
+leaf_type:;
+            if (dest_end < dest + dest_len) /* attempt to dive into file */
+                return ISO_DEAD_SYMLINK;
+            *res = n;
+            return ISO_SUCCESS;
+        }
+    }
+    *res = (IsoNode *) cur_dir;
+    return ISO_SUCCESS;
+}
+
